%
% Apply a trained classifier to new data 
%
% In
% - the output of one of the classifier functions
% - examples - #examples x #features
%
% Out
% - a vector of predicted labels - #examples x 1
% - scores - #examples x #classes - highest score in row <i> is predicted class for example
%
% Examples:
% - [predictedLabels] = applyClassifier( model, examples );
%
% History
% - 2009 October 22 - created - fpereira@princeton.edu
%

function [predictedLabels,scores] = applyClassifier( varargin )

if nargin < 2
  fprintf('syntax: applyClassifier( <trained classifier>, <test examples> )\n');return;
end

model     = varargin{1}; % trained classifier
examples  = varargin{2};
[nExamples,nFeatures] = size(examples);

% unpack information from trained classifier
classifier = model.classifier;

%
% apply classifier
%

  
trainingSetInfo = model.trainingSetInfo;
nClasses          = trainingSetInfo.nClasses;
sortedLabelValues = trainingSetInfo.sortedLabelValues;
classPriors       = trainingSetInfo.classPriors;

switch classifier
 case {'logisticRegression'}  
  scores = examples * model.W(2:end,:) + repmat(model.W(1,:),nExamples,1);
  [discard,maxpos] = max(scores,[],2);
  predictedLabels  = sortedLabelValues(maxpos);
 
 case {'libsvm'}
  if model.probabilityEstimate
    [predictedLabels,discard,probabilityEstimates] = svmpredict(zeros(nExamples,1),examples,model.libsvmModel,'-b 1');
    scores = probabilityEstimates;
  else    
    [predictedLabels] = svmpredict(zeros(nExamples,1),examples,model.libsvmModel);
    scores = zeros(nExamples,nClasses);
    switch model.kernel
     case {'linear'}  
      if nClasses == 2 
        scores = examples * model.W(2:end,:) + repmat(model.W(1,:),nExamples,1);
      end
     otherwise
    end
  end
  
 case {'bayes'}
  classifierModel = model.classifierModel;
  
  % this is a bit more elaborate
  classMean = model.classMean;
  classStdv = model.classStdv;

  scores = zeros(nExamples,nClasses);
  
  differentPriors = sum(diff(classPriors));

  switch classifierModel
   case {'gnb'}
    % as the standard deviation of each feature is the same
    % if the priors are all the same we don't need the exact log
    
    if differentPriors
      constantPart = -nFeatures*log(sqrt(2*pi))-sum(repmat(log(classStdv),nExamples,1),2);
    end
    
    for ic = 1:nClasses

      if differentPriors
        scores(:,ic) = constantPart - sum(((examples-repmat(classMean(ic,:),nExamples,1))./repmat(classStdv,nExamples,1)).^2,2)/2;
        scores(:,ic) = scores(:,ic) + log(classPriors(ic));
      else
        scores(:,ic) = - sum( (examples - repmat(classMean(ic,:),nExamples,1)).^2, 2);
      end
    end
    
    [maxval,maxpos] = max(scores,[],2);
    predictedLabels = sortedLabelValues(maxpos);
  
    % debug
    %wscores = examples * model.W(2:end,1) + model.W(1,1);
    %mask = wscores >= 0;
    %wpredictedLabels = zeros(nExamples,1);
    %wpredictedLabels( mask) = sortedLabelValues(1);
    %wpredictedLabels(~mask) = sortedLabelValues(2);
    %disp([predictedLabels,wpredictedLabels,wscores]);pause
    
   case {'lda','lda_shrinkage'}
    
    if differentPriors
      fprintf('error: not implemented yet!\n');pause;return
    else
      % precompute inverse (same for all classes)
      preinv = inv(classStdv{1});
    end
        
    for ic = 1:nClasses
      if differentPriors
      else
        tmp = examples - repmat(classMean(ic,:),nExamples,1);
        scores(:,ic) =  - sum((tmp * preinv).* tmp,2) / 2;
      end
    end
    
    [maxval,maxpos] = max(scores,[],2);
    predictedLabels = sortedLabelValues(maxpos);
  
    % debug
    %wscores = examples * model.W(2:end,1) + model.W(1,1);
    %mask = wscores >= 0;
    %wpredictedLabels = zeros(nExamples,1);
    %wpredictedLabels( mask) = sortedLabelValues(1);
    %wpredictedLabels(~mask) = sortedLabelValues(2);
    %disp([predictedLabels,wpredictedLabels,wscores]);pause
    
  end
  
end
