% A wrapper for libsvm that
%
% - can do cross-validation to set the lambda parameter
% - can return feature weights in the linear case
%
% (for now ignore cross-validating other parameters)
%
% In:
% - examples - assumes a #examples x #features array
% - labels   - #examples x 1 vector
% - optional parameters (add arguments (examples,labels,<parameterName>,<parameterValue>,...)
%   - 'kernel' - kernel type (default 'linear'), options are
%                'linear'|'quadratic'|'rbf'
%                
%   - 'lambda' - the value of the tradeoff parameter for regularization (default 1)
%                or 'crossValidation' (set by cross-validation in the training set)
%
% Out:
% - model - a model structure with a trained classifier and training set information.
%   fields of interest:
%
%   - W - a 1+#features x 2 matrix of weights (first row contains bias terms for each class)
%
%     either W(2:end,1) or W(2:end,2) can be used as vectors of weights, they have opposite signs
%
%     To classify with weights,
%
%     scores = testExamples * W(2:end,:) + repmat(W(1,:),size(testExamples,1),1);
%   
% Dependencies:
% - ExternalPackages/libsvm
%
% Notes:
% - if training a linear model this code will add two fields to the output
%
%   .weights = feature weights
%   .rho     = bias
%
%   so that you can get decision values by
%
%   decisions = examples * weights + rho
%
%  (with more than two classes you have to look at how libsvm does multi-class)
%
% History:
% - 2009 October 22  - modified output to harmonize with applyClassifier
% - 2009 September 9 - created - fpereira@princeton.edu
%
% Examples:
%
% [model] = classifierLIBSVMwrapper(examples,labels);
% [model] = classifierLIBSVMwrapper(examples,labels,'kernel','linear');
% [model] = classifierLIBSVMwrapper(examples,labels,'kernel','linear','lambda',1);
%

function [model] = classifierLIBSVMwrapper( varargin )

%
% process arguments
%

this = 'classifierLIBSVMwrapper';

if nargin < 2; eval(sprintf('help %s;',this)); return; else

  examples = varargin{1}; [nExamples,nFeatures] = size(examples);
  labels = varargin{2};

  % defaults for options
  kernel = 'linear';
  lambda = 1;
  useNeighbourhood = 0;
  voxelsToNeighbours = []; numberOfNeighbours = [];
  useSilence = 0;
  W = [];
  labelsGroup = [];
  % default for
  gamma  = 1/nFeatures; % default value and term used in libsvm docs
  sigma  = sqrt(1/(2*gamma));
  coef   = 0;           % default value and term used in libsvm docs
  degree = 2;
  crossvalidateLambda = 0;
  crossvalidateGamma  = 0;
  kernelDefined = 0;
  probabilityEstimate = 0; % train to output probability estimate
  
  if nargin > 2
    % there are additional arguments to process
  
    idx = 3;
    while idx <= nargin
      argName = varargin{idx}; idx = idx + 1;

      switch argName
       case {'probabilityEstimate'}
        probabilityEstimate = 1;
       case {'kernel'}
	kernel = varargin{idx}; idx = idx + 1;
        kernelDefined = 1;
       case {'groupLabels'}
        labelsGroup = varargin{idx}; idx = idx + 1;
       case {'lambda'}
        lambda = varargin{idx}; idx = idx + 1;
	if isequal(lambda,'crossValidation')
          crossvalidateLambda = 1;
          if (isempty(labelsGroup) | ~isrealmat(labelsGroup) | (length(labelsGroup)~=nExamples))
            fprintf('error: to select lambda with crossvalidation you need to provide group labels\n');pause;return;          
          end
        else
          if lambda <= 0; fprintf('error: lambda must be positive\n',this); pause;return; end
	end
        % parameter for kernels other than linear
       case {'degree'}
        if ~kernelDefined; fprintf('error: please specify kernel argument first\n');return;end
        degree = varargin{idx}; idx = idx + 1;
       case {'gamma'}
        if ~kernelDefined; fprintf('error: please specify kernel argument first\n');return;end
        if isequal(kernel,'linear'); fprintf('error: no gamma with a linear kernel\n');end
        gamma = varargin{idx}; idx = idx + 1;
        if isequal(gamma,'crossValidation')
          crossvalidateGamma = 1;
          if (isempty(labelsGroup) | ~isrealmat(labelsGroup) | (length(labelsGroup)~=nExamples))
            fprintf('error: to select gamma with crossvalidation you need to provide group labels\n');pause;return;          
          end
        else
          switch kernel
           case {'rbf'}
            sigma = sqrt(1/(2*gamma));
           otherwise
            % nothing else to derive
          end
        end
       case {'sigma'}
        if ~kernelDefined; fprintf('error: please specify kernel argument first\n');return;end
        sigma = varargin{idx}; idx = idx + 1;
        switch kernel
         case {'rbf'}
          gamma = 1/(2*sigma^2);
         otherwise
          fprintf('error: sigma argument only works in RBF kernels\n');return;
        end
       case {'coef'}
        if ~kernelDefined; fprintf('error: please specify kernel argument first\n');return;end
        coef = varargin{idx}; idx = idx + 1;
       
       case {'neighbourhood'}
	useNeighbourhood = 1;
	voxelsToNeighbours = varargin{idx}; idx = idx + 1;
	numberOfNeighbours = varargin{idx}; idx = idx + 1;
       case {'useSilence'}
        useSilence = 1;
       otherwise
	fprintf('%s: unknown argument %s\n',this,argName); pause;return
      end 
    end
  end
  degree = round(degree);
  
  % defining a kernel argument string simplifies the rest of the code,
  % also define the parameter string for non-crossvalidated parameters
  
  parameterString = '';
  if ~crossvalidateLambda
    parameterString = sprintf('%s -c %s',parameterString,num2str(lambda));
  end
  if ~crossvalidateGamma & ~isequal(kernel,'linear')
    parameterString = sprintf('%s -g %s',parameterString,num2str(gamma));
  end
  
  % coefficient r is not cross-validated so might as well put it in kernelString
  switch kernel
   case {'linear'}
    kernelString = sprintf('-t 0 ');
   case {'quadratic','polynomial'}
    if isequal(kernel,'quadratic'); degree = 2; end
    kernelString = sprintf('-t 1 -d %d',degree);
    if coef > 0; kernelString = sprintf('%s -r %s',kernelString,num2str(coef)); end
   case {'rbf'}
    kernelString = sprintf('-t 2 ');
    if coef > 0; kernelString = sprintf('%s -r %s',kernelString,num2str(coef)); end
   case {'sigmoid'}
    kernelString = sprintf('-t 3 ');
    if coef > 0; kernelString = sprintf('%s -r %s',kernelString,num2str(coef)); end
   otherwise
    fprintf('%s: unknown kernel type %s\n',this,kernel); pause;return;
  end
end

if probabilityEstimate
  kernelString = sprintf('%s -b 1',kernelString);
end

%% compute information

sortedLabelValues     = sort(unique(labels)); conditions = sortedLabelValues;
nClasses              = length(sortedLabelValues);

%% find examples in each class and create a binary indicator mask

delta = zeros(nExamples,nClasses);

for c = 1:nClasses
  label  = sortedLabelValues(c);  
  indices{c}  = find(labels == label);
  nperclass(c) = length(indices{c});

  delta(indices{c},c) = 1;
end


%% compute class mean examples

meanExamples = zeros(nClasses,nFeatures);
meanLabels   = zeros(nClasses,nClasses);

for c = 1:nClasses
  meanExamples(c,:) = mean(examples(indices{c},:),1);
  meanLabels(c,1)   = sortedLabelValues(c);
end  


%% if requested, run a cross validation to set the value of lambda

if crossvalidateLambda | crossvalidateGamma
  fprintf('%s: using crossvalidation to find lambda\n',this);
  
  groups = unique(labelsGroup); nGroups = length(groups);
  
  if crossvalidateLambda
    lambdaRange = [1 10 0.1 100 0.01 1000 0.001];
  else
    lambdaRange = [lambda]; % fake range for outer loop
  end
  nl = length(lambdaRange);
    
  if crossvalidateGamma
    % range to try depends on the number of features
    gammaRange = [1/nFeatures, 10.^(ceil(log10(1/nFeatures)):log10(1))]
  else
    switch kernel
     case {'linear'}
      gammaRange = [0];
     otherwise
      gammaRange = [gamma]; % fake range for outer loop
    end
  end
  ng = length(gammaRange);
    
  cvresults = zeros(nl,ng);
  
  for ir = 1:nl
    lambda = lambdaRange(ir);
    
    baseString = sprintf(' -c %s',num2str(lambda));
    
    fprintf('\ttesting lambda=%s\n',num2str(lambda));

    for ig = 1:ng
      gamma = gammaRange(ig);
      
      if gamma
        parameterString = sprintf('%s -g %s',baseString,num2str(gamma));
      else
        % in linear kernel
        parameterString = baseString;
      end
      completeString  = sprintf('%s %s',kernelString,parameterString);
      
      fprintf('\t\ttesting gamma=%s\n',num2str(gamma));
      
      predictedLabels = zeros(nExamples,1);
    
      for g = 1:nGroups
        group = groups(g);
            
        mask = (labelsGroup == group);

        %fprintf('\t\tcrossvalidation loop\n');
        indicesTest  = find( mask); nTest = length(indicesTest);
        indicesTrain = find(~mask); 
      
        cvModel = svmtrain(labels(indicesTrain),examples(indicesTrain,:),completeString);
        predictedLabels(indicesTest,:) = svmpredict(zeros(nTest,1),examples(indicesTest,:),cvModel);
      end

      cvresults(ir,ig) = sum(predictedLabels~=labels)/nExamples; % error
    
    end; % loop over gamma
  end; % loop over lambda

  [minval,minposgamma]  = min(cvresults,[],2);
  [minval,minposlambda] = min(minval); minposgamma = minposgamma(minposlambda);

%  cvresults
%  minposgamma
%  minposlambda
  
  lambda = lambdaRange(minposlambda);
  gamma  = gammaRange(minposgamma);
  
  fprintf('%s: crossvalidation lambda/gamma are %s/%s\n',this,num2str(lambda),num2str(gamma));
  
  clear examplesTrain examplesTest labelsTrain labelsTest;
end


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% Train classifier
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

switch kernel
 case {'linear'}
  completeString = sprintf('%s -c %s',kernelString,num2str(lambda));
 otherwise
  completeString = sprintf('%s -c %s -g %s',kernelString,num2str(lambda),num2str(gamma));
end

% this is it, really
libsvmModel = svmtrain(labels,examples,completeString);

% derive other things from the trained model
switch kernel
 case {'linear'}
  if nClasses == 2
    % get weights
    W = zeros(nFeatures+1,nClasses);
  
    weights = full(sum(libsvmModel.SVs .* repmat(libsvmModel.sv_coef,1,nFeatures)))';
    rho = -1 * libsvmModel.rho;
  
    W(1,:)     = [rho -rho];
    W(2:end,:) = [weights -weights]; 

    if 0
      % debug stuff
      
      decisionsWeights = examples*weights + rho;
      
      % apply model to training data to check
      [discard1,discard2,decisionsModel] = svmpredict(zeros(size(labels)),X,model.libsvmModel);
      
      scores = zeros(nExamples,2);
      % inefficient to check every time, but shouldn't be too bad
      if model.libsvmModel.Label(1) < model.libsvmModel.Label(2)
        % labels are in the right order
        scores(:,1) = decisionsModel;
        scores(:,2) = -decisionsModel;
      else
        % labels are flipped
        scores(:,1) = -decisionsModel;
        scores(:,2) = decisionsModel;
      end
      scoresWeights = [ones(nExamples,1),X] * model.W;
      disp([decisionsWeights,decisionsModel,scores,scoresWeights]);
      pause
    end
  else
    W = [];
    % no support yet for multiclass weights
  end
    
 otherwise
  % no weights to learn
  
end


%% Pack the results into a struct

model.classifier = 'libsvm';
model.libsvmModel = libsvmModel;
model.probabilityEstimate = probabilityEstimate;
switch kernel
 case {'linear'}
  if nClasses == 2
    model.W = W;
    model.Wbias       = W(1,:);     % to help people
    model.Wfeatures   = W(2:end,:); % decipher the output
  end
 otherwise
end

% Training Set information
trainingSetInfo.nExamples         = nExamples;
trainingSetInfo.nFeatures         = nFeatures;
trainingSetInfo.nLabels           = nClasses;
trainingSetInfo.nClasses          = nClasses; % same thing as labels
trainingSetInfo.sortedLabelValues = sortedLabelValues;
trainingSetInfo.classPriors       = nperclass / sum(nperclass);
model.trainingSetInfo = trainingSetInfo;
model.kernel = kernel;
switch kernel
 case {'linear'}
  model.lambda = lambda;
 otherwise
  model.lambda = lambda;
  model.gamma  = gamma;
end

return;


function [] = test()

nPerClass = 100; 
n = nPerClass*2;
m = 100; 

tmp = repmat(1:10,10,1);
labelsGroup = [tmp(:);tmp(:)];

examplesA = randn(nPerClass,m);
examplesB = randn(nPerClass,m);
examplesB(:,1:(m/2)) = examplesB(:,1:(m/2)) + 0.5;
examples = [examplesA;examplesB];
labels = [repmat(1,nPerClass,1);repmat(2,nPerClass,1)];

[trainedClassifier] = classifierLIBSVMwrapper(examples,labels);
[trainedClassifier] = classifierLIBSVMwrapper(examples,labels,'groupLabels',labelsGroup,'lambda','crossvalidation');

kernel = 'quadratic';
kernel = 'polynomial';
kernel = 'rbf';
kernel = 'sigmoid';

[trainedClassifier] = classifierLIBSVMwrapper(examples,labels,'groupLabels',labelsGroup,'kernel',kernel,'lambda','crossValidation','gamma','crossValidation');



examplesA = randn(nPerClass,m);
examplesB = randn(nPerClass,m);
examplesB(:,1:(m/2)) = examplesB(:,1:(m/2)) + 0.5;
examplesTest = [examplesA;examplesB];
labelsTest = [repmat(1,nPerClass,1);repmat(2,nPerClass,1)];
nTest = 2*nPerClass;

[predictedLabels] = applyClassifier(trainedClassifier,examples);

sum(predictedLabels==labelsTest)/n
